// Description: a node that allows drawing to one or both eyes
//
//<b>notes:</b>
//<ul>
//<li> all parent node messages are inherited
//<li> see reset method for default settings
//<li> you must add the following to your .ygrc file in order to use this node:
//<ul>
//<li> setenv YG_PRELOAD_CLASSES CAVEEyes
//</ul>
//</ul>
//
// Category: Attributes
// Author: Alex Hill
// Revision: 04/18/05
//
#include <Performer/pf/pfGroup.h>
#include <ygNetMessages.h>
#include <cave.h>
#include "CAVEEyes.h"

using namespace std;

extern "C" ygNode* construct_CAVEEyes(const char* name,bool master) { return new CAVEEyes(name,master); }

typedef struct _eyeData
    {
    int eyes;
    } eyeData_t;

 
CAVEEyes::CAVEEyes(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("CAVEEyes",true);
    //allocate shared memory eyeData structure for pre and post callbacks
    data_ = (eyeData_t*) pfMalloc(sizeof(eyeData_t),pfGetSharedArena());
    data_->eyes = CAVE_BOTH_EYES;
    pfnode()->setTravFuncs(PFTRAV_CULL,eyePreCull,NULL);
    pfnode()->setTravData(PFTRAV_CULL,data_);
    //distribute the name of the node affected
    addNetMessage("eyes",&data_->eyes,YG_NET_INT);
    addNetMessageMap("eyes","left",CAVE_LEFT_EYE);
    addNetMessageMap("eyes","right",CAVE_RIGHT_EYE);
    addNetMessageMap("eyes","both",CAVE_BOTH_EYES);
    }


CAVEEyes::~CAVEEyes(void)
    {
    pfDelete(data_);
    }


void CAVEEyes::reset(void)
    {
    resetNetMessage("eyes");
    ygNode::reset();
    ygString args("interocular");
    char val[32];
    args += "=";
    sprintf(val,"%f",CAVEConfig->InterocularDistance);
    args += val;
    //generate changed event with interocular distance
    eventOccurred("changed",args);
    }


void CAVEEyes::message(const ygMessage& msg)
    {
    //eye mode to allow rendering
    if (msg == "eyes")
        {
        return;
        }
    else
        ygNode::message(msg);
    }

    
int CAVEEyes::eyePreCull(pfTraverser*,void *data)
    {
    //cast the traversal data as the eyeData structure
    eyeData_t* eyeData = (eyeData_t*)data;
    //if the plane is active then
    if (eyeData->eyes == CAVE_LEFT_EYE && CAVEEye != CAVE_LEFT_EYE)
        return PFTRAV_PRUNE;
    if (eyeData->eyes == CAVE_RIGHT_EYE && CAVEEye != CAVE_RIGHT_EYE)
        return PFTRAV_PRUNE;
    return PFTRAV_CONT;
    }
