// Description: generates events describing node event names
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//</ul>
//
// Category: System
// Author: Alex Hill
// Revision: 08/15/05 Alex Hill
//
#include <vector>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include "eventListing.h"

using namespace std;

extern "C" ygNode* construct_eventListing(const char* name,bool master) { return new eventListing(name,master); }

eventListing::eventListing(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("eventListing");
    index = 0;
    maximum = 100;
    node = NULL;
    addNetMessage("node",&nodeName,YG_NET_NODE);
    addNetMessage("events",&eventName,YG_NET_STRING);
    addNetMessage("events",&index,YG_NET_INT);
    addNetMessage("maximum",&maximum,YG_NET_INT);
    addEvent("events");
    addEvent("changed");
    }
    
void eventListing::reset(void)
    {
    node = NULL;
    lastNode.clear();
    lastEvent.clear();
    lastIndex = -1;
    //clear the node name
    resetNetMessage("node");
    //clear the event name
    resetNetMessage("events");
    //set the maximum printout to 100
    resetNetMessage("maximum");
    ygNode::reset();
    }

void eventListing::message(const ygMessage& msg)
    {
    //generate events with event names
    if (msg == "node")
        {
        return;
        }
    //generate events with event message names
    else if (msg == "events")
        {
        checkEvents();
        }
    //generate events with event message names
    else if (msg == "index")
        {
        if (msg.numArgs() > 0)
            {
            index = msg.intArg(0);
            lastIndex = -1;
            checkEvents();
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //set the total number of directories and files to report
    else if (msg == "maximum")
        {
        return;
        }
    else
        ygNode::message(msg);
    }


void eventListing::checkEvents(void)
    {
    if (nodeName != lastNode)
        {
        if (nodeName.length() > 0)
            {
            node = ygNodeDB::find(nodeName,false);
            if (node)
                lastIndex = -1;
            }
        else
            node = NULL;
        lastNode = nodeName;
        }
    if (eventName != lastEvent)
        {
        if (node)
            lastIndex = -1;
        lastEvent = eventName;
        }
    if (node && index != lastIndex)
        {
        bool found=false;
        const map<ygString,vector<ygString> >& events = node->getEventMessages();
        map<ygString,vector<ygString> >::const_iterator i;
        for (i=events.begin(); i != events.end(); ++i)
            {
            const vector<ygString>& message = (*i).second;
            if ((*i).first == eventName)
                {
                found = true;
                int messageCount = message.size();
                int messageMax = index + maximum;
                if (messageMax > messageCount)
                    messageMax = messageCount;
                ygString args("total=");
                char val[32];
                sprintf(val,"%d",messageCount);
                args += val;
                eventOccurred("changed",args);
                int num=0;
                for (int i=index;i<messageMax;i++)
                    {
                    ygString args("data=\"");
                    args += message[i];
                    args += "\" number=";
                    sprintf(val,"%d",num);
                    args += val;
                    args += " index=";
                    sprintf(val,"%d",i);
                    args += val;
                    eventOccurred("events",args);
                    num++;
                    }
                }
            }
        if (!found)
            {
            ygString args("total=0");
            eventOccurred("changed",args);
            }
        }
    lastIndex = index;
    }
