// particleSystem.h : header file
// Note: implemented by S. James Lee 2003.11.

#ifndef _particleSystem_h_
#define _PArticleSystem_h_

#include "ygNode.h"
#include "ygTransform.h"
#include <string>
#include <Performer/pf/pfDCS.h>

using namespace std;

struct	_particleSystemPrivateData;

class particleSystem;

class Particle
{
public:
	Particle();
	~Particle();
	
	particleSystem*	m_pParentSystem;
	
	pfVec3	m_vPrevPosition;
	pfVec3	m_vPosition;
	pfVec3	m_vVelocity;
	pfVec4	m_Color;
	pfVec4	m_ColorDelta;
	float	m_fGravity;
	float	m_fAge;
	float	m_fLifetime;
	float	m_fSize;
	float	m_fSizeDelta;
	float	m_fRoll;
	float	m_fRollDelta;
	float	m_fAlpha;
	float	m_fAlphaDelta;
	float	m_fGap;
	
	bool Update(float fTimeDelta);
	void UpdateTransform();
	void initGeometry(int iBlendType);
	void setBlend(int iBlendType);	
	void Restart();
		
	pfVec3*	m_pVertices;
	pfVec4*	m_pColor;
	pfVec3*	m_pNormal;
	pfVec2*	m_pTCoord;
	pfTexture* m_pTexture;
	pfTexEnv *m_pTev;
	
	// performer object
	pfGeode* geode;
	pfGeoSet* geoset;
	pfGeoState* gstate;
	pfVec3 axis;
	pfDCS* dcs;
};
 
/*************************************************************************/
// particleSystem class
/*************************************************************************/
class particleSystem : public ygNode
{
public:
	particleSystem(const char* name, bool master=true);
	~particleSystem(void);

	virtual void reset(void);
	virtual void message(const ygMessage&);
	virtual void app(void);
	
	virtual bool IsColliding();
	virtual bool IsAttractive();


	void KillThemAll();
	bool LoadSFXFile(const ygString& file);
	void LoadChainSystem(string& str);
	void ResetPartlceArray();
	void Restart();
	int  NumParticles();
	void Move(pfVec3 vPosition);
	
	bool Update(float fTime);	
	bool UpdateTransform();
	void GetLocation(pfVec3& vLocation);
	void InitSystem();
	void setBlend(int iBlendType);

	_particleSystemPrivateData * p_;
	  
};

#endif
