// Description: an array that is created and updated on each client
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//</ul>
//
// Category: Math
// Author: Alex Hill
// Revision: 01/21/03
//
#include <ygWorld.h>
#include <ygNetMessages.h>
#include "remoteValue.h"

#define VALUE_NETKEY "v"

extern "C" ygNode* construct_remoteValue(const char* name,bool master) { return new remoteValue(name,master); }
char* remoteValue_dependencies = "YG DSO Dependencies: value";

remoteValue::remoteValue(const char* name,bool master) : value(name,master)
    {
    setClassName("remoteValue",true);
    //distribute the size of the array
    addNetMessage(VALUE_NETKEY,&values,YG_NET_VECTOR);
    //dont include in documentation
    dontSaveNetMessage(VALUE_NETKEY);
    updateInterval = 0.0;
    lastUpdate = 0.0;
    addNetMessage("updateInterval",&updateInterval,YG_NET_FLOAT);
    }

remoteValue::~remoteValue(void)
    {
    }

void remoteValue::reset(void)
    {
    //reset update interval
    resetNetMessage("updateInterval");
    //reset last update timer
    lastUpdate = 0;
    value::reset();
    }

void remoteValue::message(const ygMessage& msg)
    {
    //set the values of all array elements
    if (msg == VALUE_NETKEY)
        {
        if (values.size() != size)
            setValueSize(values.size());
        }
    else
        value::message(msg);
    }
    
void remoteValue::app(void)
    {
    value::app();
    if (isNetMaster())
        {
        //if update interval has passed then update key
        if ((ygWorld::frameTime() - lastUpdate) > updateInterval)
            {
            bool changed=false;
            for (int i=0;i<size;i++)
                {
                if (values[i] != netValues[i])
                    changed = true;
                }
            if (changed)
                {
                netValues = values;
                netMessageChanged(VALUE_NETKEY);
                lastUpdate = ygWorld::frameTime();
                }
            }
        }
    }

void remoteValue::setValueSize(int newSize)
    {
    setSize(values,newSize,values.size());
    setSize(lastValues,newSize,lastValues.size());
    setSize(netValues,newSize,netValues.size());
    size = newSize;
    }

