#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <QUANTA/QUANTA.hxx>
#include "ygNetClient.h"
#include "ygKeyDatabase.h"
#include "ygrepeaterQueue.h"

using namespace std;


#define MAX_LINKS 256

typedef struct
    {
    char* linkHost[MAX_LINKS];
    int numLinks;
    int portNumber;
    int queueSize;
    int cachePortNumber;
    int statsInterval, dumpInterval;
    } RepeaterOpts_t;


void parseCommandLine(int argc,char** argv,RepeaterOpts_t* opts);
void* udpLoop(void*);
void* tcpLoop(void*);
void* databaseCache(ygrepeaterQueue*,ygrepeaterQueue*);


RepeaterOpts_t repOpts;


main(int argc, char **argv)
    {
    cout << "ygrepeater for Ygdrasil version 0.4.3\n";
    repOpts.numLinks = 0;
    repOpts.portNumber = 3500;
    repOpts.queueSize = 2000000;
    repOpts.cachePortNumber = 0;
    repOpts.statsInterval = 0;
    repOpts.dumpInterval = 0;
    parseCommandLine(argc,argv,&repOpts);
    if (!repOpts.cachePortNumber)
        repOpts.cachePortNumber = repOpts.portNumber+1;
    QUANTAinit();
    ygrepeaterQueue* udpQueue = new ygrepeaterQueue(repOpts.queueSize);
    ygrepeaterQueue* tcpQueue = new ygrepeaterQueue(repOpts.queueSize);
    QUANTAts_thread_c udpThread;
    udpThread.create(udpLoop, udpQueue);
    QUANTAts_thread_c tcpThread;
    tcpThread.create(tcpLoop, tcpQueue);
    databaseCache(tcpQueue,udpQueue);
    }


void parseCommandLine(int argc,char **argv,RepeaterOpts_t *opts)
    {
    int i;
    for (i=1; i < argc; i++)
        {
        if (!strncmp(argv[i],"--",2))
            {
            if (!strcasecmp(argv[i],"--link"))
                {
                i++;
                if (opts->numLinks < MAX_LINKS)
                    {
                    opts->linkHost[opts->numLinks] = strdup(argv[i]);
                    opts->numLinks++;
                    }
                else
                    fprintf(stderr, "ERROR: cannot link to more than"
                        " %d hosts; link to '%s' ignored\n",
                        MAX_LINKS, argv[i]);
                }
            else if (!strcasecmp(argv[i],"--port"))
                {
                i++;
                opts->portNumber = atoi(argv[i]);
                }
            else if (!strcasecmp(argv[i],"--queuesize"))
                {
                i++;
                opts->queueSize = atoi(argv[i]);
                }
            else if (!strcasecmp(argv[i],"--cacheport"))
                {
                i++;
                opts->cachePortNumber = atoi(argv[i]);
                }
            else if (!strcasecmp(argv[i],"--stats"))
                {
                i++;
                opts->statsInterval = atoi(argv[i]);
                }
            else if (!strcasecmp(argv[i],"--dump"))
                {
                i++;
                opts->dumpInterval = atoi(argv[i]);
                }
            else
                fprintf(stderr,"WARNING: unknown command line "
                    "option '%s'\n",argv[i]);
            }
        }
    }


/**************************************************************************/

int udpIntercept(QUANTAnet_udp_c *,char**,int *,void *);
int tcpIntercept(QUANTAnet_tcpReflectorClient_c *client, char** buffer,
        int *bufferSize, void *userData);

void* udpLoop(void *userData)
    {
    int i, sprevTime=time(NULL), nowTime;

    ygrepeaterQueue *repqueue = (ygrepeaterQueue *) userData;
    QUANTAnet_udpReflector_c *udpReflector;
    udpReflector = new QUANTAnet_udpReflector_c;
    udpReflector->setIncomingPort(repOpts.portNumber);
    for (i=0; i < repOpts.numLinks; i++)
        udpReflector->setForcedDestination(repOpts.linkHost[i], repOpts.portNumber);
    udpReflector->intercept(udpIntercept,repqueue);
    udpReflector->init();

    signal(SIGSEGV, SIG_DFL);
    signal(SIGBUS, SIG_DFL);
    while (true)
        {
        udpReflector->process();
        if (repOpts.statsInterval > 0)
            {
            nowTime = time(NULL);
            if (nowTime - sprevTime >= repOpts.statsInterval)
                {
                udpReflector->showStats("UDPReflector","ygrepeater043 UDPReflector");
                sprevTime = nowTime;
                }
            }
        }
    return NULL;
    }


void* tcpLoop(void *userData)
    {
    int i, sprevTime=time(NULL), nowTime;

    ygrepeaterQueue *repqueue = (ygrepeaterQueue *) userData;
    QUANTAnet_tcpReflector_c *tcpReflector;
    tcpReflector = new QUANTAnet_tcpReflector_c;
    tcpReflector->init(repOpts.portNumber);
    for (i=0; i < repOpts.numLinks; i++)
        tcpReflector->setForcedDestination(repOpts.linkHost[i], repOpts.portNumber);
    tcpReflector->intercept(tcpIntercept,repqueue);

    signal(SIGSEGV, SIG_DFL);
    signal(SIGBUS, SIG_DFL);
    while (true)
        {
        tcpReflector->process();
        if (repOpts.statsInterval > 0)
            {
            nowTime = time(NULL);
            if (nowTime - sprevTime >= repOpts.statsInterval)
                {
                tcpReflector->showStats("TCPReflector","ygrepeater043 TCPReflector");
                sprevTime = nowTime;
                }
            }
        }
    return NULL;
    }


int udpIntercept(QUANTAnet_udp_c *client, char** buffer, int *bufferSize,
        void *userData)
    {
    ygrepeaterQueue *repqueue = (ygrepeaterQueue *) userData;
    repqueue->add(*buffer,*bufferSize);
// Previously, there were 8 garbage bytes on the front of UDP packets;
// now, they seem to be okay.
// ???
//    repqueue->add((*buffer)+8,(*bufferSize)-8);
    return QUANTAnet_udpReflector_c::OK;
    }


int tcpIntercept(QUANTAnet_tcpReflectorClient_c *client, char** buffer,
        int *bufferSize, void *userData)
    {
    ygrepeaterQueue *repqueue = (ygrepeaterQueue *) userData;
    repqueue->add(*buffer,*bufferSize);
    return QUANTAnet_tcpReflector_c::OK;
    }


/**************************************************************************/
/**************************************************************************/


/** Note:  I would have just used a QUANTAnet_tcpServer for the cache's
   TCP connection, since we don't need to reflect data among clients.
   But, because QUANTAnet_tcpReflector handles the packetizing of data
   automatically, and because the other, existing code currently expects
   a QUANTAnet_tcpReflectorClient, it was simpler to use the reflector,
   and just prevent it from doing any actual reflection.
   Clients will be expected to disconnect once the download is finished.
**/

int cacheTCPClient(QUANTAnet_tcpReflectorClient_c *client, char** buffer,
        int *bufferSize, void *userData);
void checkRepeaterQueue(ygrepeaterQueue *repqueue,ygKeyDatabase *ygdb);
void saveKeyData(ygKeyDatabase *db,QUANTAnet_datapack_c& packer);


void* databaseCache(ygrepeaterQueue* primary,ygrepeaterQueue* secondary)
    {
    FILE *dumpFP;
    int prevTime=time(NULL), nowTime;
    struct timespec sleeptime;
    ygKeyDatabase *ygdb = new ygKeyDatabase;
    QUANTAnet_tcpReflector_c *server = new QUANTAnet_tcpReflector_c;
    server->init(repOpts.cachePortNumber);
    server->intercept(cacheTCPClient,ygdb);
    if (repOpts.dumpInterval > 0)
        {
        char filename[256];
        sprintf(filename, "ygdbdump.%d", getpid());
        dumpFP = fopen(filename, "w");
        if (!dumpFP)
            dumpFP = stdout;
        }
    while (true)
        {
        server->process();
        checkRepeaterQueue(primary,ygdb);
        checkRepeaterQueue(secondary,ygdb);
        if (repOpts.dumpInterval > 0)
            {
            nowTime = time(NULL);
            if (nowTime - prevTime >= repOpts.dumpInterval)
                {
                ygdb->printContents(dumpFP);
                prevTime = nowTime;
                }
            }
/* Sleep since we can't block on a select(); this thread doesn't need to be as
   ultra-responsive as the reflector thread */
        sleeptime.tv_sec = 0;
        sleeptime.tv_nsec = 1000000;
        nanosleep(&sleeptime,NULL);
        }
    return NULL;
    }


int cacheTCPClient(QUANTAnet_tcpReflectorClient_c *client, char** buffer,
        int *bufferSize, void *userData)
    {
    ygKeyDatabase *ygdb = (ygKeyDatabase *) userData;
    QUANTAnet_datapack_c packer;
    int packetType;
    packer.initUnpack(*buffer,*bufferSize);
    packer.unpackInt(&packetType);
    if (packetType == YG_NET_REQUEST_ALL)
        {
        char clientName[256];
        client->getRemoteIP(clientName);
        printf("sending database to new client %s\n",clientName);
        ygdb->sendContents(client);
        printf("done sending database\n");
        fflush(stdout);
        }
    return QUANTAnet_udpReflector_c::SKIP_DISTRIBUTION;
    }


void checkRepeaterQueue(ygrepeaterQueue *repqueue,ygKeyDatabase *ygdb)
    {
    char buffer[YG_NET_MAX_BUFFER_SIZE];
    size_t packetSize;
    while (repqueue->get(buffer,&packetSize,sizeof(buffer)) > 0)
        {
        QUANTAnet_datapack_c packer;
        int packetType;
        packer.initUnpack(buffer,packetSize);
        while (true)
            {
            int packetType = -1;
            if (packer.unpackInt(&packetType) ==
                QUANTAnet_datapack_c::FAILED)
                break;
            if (packetType == YG_NET_PUT_KEY)
                saveKeyData(ygdb, packer);
            else
                break;
            }
        }
    }


void saveKeyData(ygKeyDatabase *db,QUANTAnet_datapack_c& packer)
    {
    static char buffer[YG_NET_MAX_BUFFER_SIZE];
    ygString path, key;
    int length;
/* Get path string length & data */
    packer.unpackInt(&length);
    packer.unpack(buffer, length);
    path.append(buffer);
/* Get key string length & data */
    packer.unpackInt(&length);
    packer.unpack(buffer, length);
    key.append(buffer);
/* Get data length & data */
    packer.unpackInt(&length);
    packer.unpack(buffer, length);
    db->store(path, key, buffer, length);
    }


/**************************************************************************/
