// Author: Dave Pape
//           11/01/01
// Revision: 02/01/06 Alex Hill - added next event logic, debug messaging, and accessor functions
//
#include <stdlib.h>
#include <vector>
#include <map>
#include "ygWorld.h"
#include "ygEvent.h"
#include "ygMessage.h"
#include "ygUtil.h"
#include "ygEventMessages.h"

using namespace std;


struct _ygEventMessagesPrivateData
    {
    map<ygString,vector<ygString> > eventMap; 
    vector<ygString> events; 
    };


struct _ygEventMessagesGlobalData
    {
    ygString nextEvent;   
    };

struct _ygEventMessagesGlobalData* ygEventMessages::global_ = NULL;


ygEventMessages::ygEventMessages(void)
    {
    if (!global_)
        {
        global_ = new struct _ygEventMessagesGlobalData;
        global_->nextEvent = "event1";
        }
    p_ = new struct _ygEventMessagesPrivateData;
    ygnode_ = NULL;
    }


ygEventMessages::~ygEventMessages(void)
    {
    delete p_;
    }


void ygEventMessages::setNode(class ygNode* node)
    {
    ygnode_ = node;
    }


void ygEventMessages::clear(void)
    {
    map<ygString,vector<ygString> >::iterator iter;
    for (iter=p_->eventMap.begin(); iter != p_->eventMap.end(); ++iter)
        (*iter).second.clear();
    p_->eventMap.clear();
    }


void ygEventMessages::clear(const ygString& event)
    {
    vector<ygString>& mapEntry = p_->eventMap[event];
    mapEntry.clear();
    }


void ygEventMessages::clear(const ygString& event,int num)
    {
    vector<ygString>& mapEntry = p_->eventMap[event];
    vector<ygString>::iterator iter = mapEntry.begin();
    int j=0;
    while ((iter != mapEntry.end()) && (j != num))
        {
        ++iter;
        ++j;
        }
    if (iter != mapEntry.end())
        mapEntry.erase(iter);
    }


void ygEventMessages::addEvent(const ygString& event)
    {
    p_->events.push_back(event);
    }


void ygEventMessages::addMessage(const ygString& event,const ygString& message)
    {
    p_->eventMap[event].push_back(message);
    if (event == global_->nextEvent)
        {
        p_->events.push_back(event);
        ygIncrementName(global_->nextEvent);
        }
    }


void ygEventMessages::respondToEvent(ygEvent* event) const
    {
    vector<ygString>::const_iterator iter;
    vector<ygString>& mapEntry = p_->eventMap[event->name()];
    for (iter = mapEntry.begin(); iter != mapEntry.end(); ++iter)
        {
        ygMessage msg;
        msg.parseString(*iter,ygnode_,event);
        if (ygnode_)
            {
            bool* flag;
            flag = ygnode_->getDebug("messages");
            if (*flag)
                {
                ygString tmp;
                msg.writeString(tmp);
                cout << ygnode_->name() << " sending message '" << tmp << "'\n";
                }
            }
        ygWorld::world()->scheduleMessage(msg);
        }
    }


const map<ygString,vector<ygString> >& ygEventMessages::getMessages(void)
    {
    return p_->eventMap;
    }


const vector<ygString>& ygEventMessages::getEvents(void)
    {
    return p_->events;
    }


const ygString& ygEventMessages::getNextEvent(void)
    {
    return global_->nextEvent;
    }
