// Author: Dave Pape
//           11/01/01
// Revision: 02/01/06 Alex Hill - parser now calls reset on each node and does name replacement
//
#include <stdio.h>
#include <iostream>
#include <stack>
#include <list>
#include "ygTokenStream.h"
#include "ygNodeDB.h"
#include "ygNode.h"
#include "ygParser.h"

using namespace std;

ygParser::ygParser(void)
    {
    }

void ygParser::clear(void)
    {
    nodeList.clear();
    replaceMap.clear();
    }

void ygParser::parseFile(const ygString& file,ygNode *root,bool preprocess)
    {
    ygTokenStream *input;
    ygString token;
    stack<ygNode*> parentStack;
    ygNode *parent=root, *lastNode=NULL;
    input = new ygTokenStream(file,preprocess);
    while (input->next(token))
        {
        if (token == "{")
            {
            parentStack.push(parent);
            if (!lastNode)
                {
                ygString name;
                ygNodeDB::createValidName("group",name);
                lastNode = ygNodeDB::create("group",name,input);
                nodeList.insert(nodeList.end(),lastNode);
                parent->addChild(lastNode);
                ygString* comment;
                while(comment = input->nextComment())
                    lastNode->addComment(comment);
                }
            parent = lastNode;
            lastNode = NULL;
            }
        else if (token == "}")
            {
            if (!parentStack.empty())
                {
                parent = parentStack.top();
                parentStack.pop();
                }
            }
        else if (token[0] == '(')
            {
            cerr << "ERROR: parsing " << input->filename()
                << ", line " << input->linenumber()
                << ": there's no node for this option list:"
                << endl << "\t" << token << endl;
            }
        else
            {
            lastNode = parseNode(token,input);
            if (lastNode)
                {
                nodeList.insert(nodeList.end(),lastNode);
                parent->addChild(lastNode);
                ygString* comment;
                while(comment = input->nextComment())
                    lastNode->addComment(comment);
                }
            }
        }
    delete input;
    while (!nodeList.empty())
        {
        ygNode* n = nodeList.front();
        replaceInitMessages(n);
        n->reset();
        nodeList.pop_front();
        }        
    }


ygNode * ygParser::parseNode(const ygString& nodeType,ygTokenStream *input)
    {
    bool gotToken=true;
    ygString token, name;
    ygNode *node;
    gotToken = input->next(token);
    if ((gotToken) && (token[0] != '(') && (token[0] != '{') &&
        (token[0] != '}'))
        {
        name = token;
        if(!ygNodeDB::createValidName(nodeType,name))
            replaceMap[token] = name;
        gotToken = input->next(token);
        }
    else
        ygNodeDB::createValidName(nodeType,name);
    node = ygNodeDB::create(nodeType,name,input);
    if (gotToken)
        {
        if (token[0] == '(')
            {
            if (node)
                parseInitMessages(node,token);
            }
        else
            input->pushback(token);
        }
    return node;
    }


void ygParser::parseInitMessages(ygNode * node,ygString& msgList)
    {
    ygString msg;
    int index=1;
    if (msgList[msgList.length()-1] != ')')
            msgList += ')';
    while (1)
        {
        msg.clear();
        while ((msgList[index] == ' ') || (msgList[index] == '\t'))
            index++;
        if (msgList[index] != ')')
            {
            int parenDepth=0, quoting=0;
            while ((parenDepth) || (quoting) ||
                ((msgList[index] != ',') &&
                 (msgList[index] != ')')))
                {
                if (msgList[index] == '\\')
                    {
                    if (quoting)
                        msg += msgList[index];
                    index++;
                    }
                else if (msgList[index] == '"')
                    quoting = !quoting;
                else if ((!quoting) && (msgList[index] == '('))
                    parenDepth++;
                else if ((!quoting) && (msgList[index] == ')'))
                    parenDepth--;
                msg += msgList[index];
                index++;
                }
            node->addInitMessage(msg);
            if (msgList[index] == ',')
                index++;
            }
        else
            break;
        }
    }


void ygParser::replaceInitMessages(ygNode* node)
    {
    const vector<ygMessage*>& messages = node->getInitMessages();
    vector<ygMessage*>::const_iterator iter;
    for (iter=messages.begin(); iter != messages.end(); ++iter)
        replaceMessageName(*iter);
    }

void ygParser::replaceMessageName(ygMessage* msg)
    {
    map<ygString,ygString>::iterator iter;
    for (iter=replaceMap.begin(); iter != replaceMap.end(); ++iter)
	msg->replaceNodeName((*iter).first,(*iter).second);
    }
